#  tests for thinc-8.0.1-py37h1341a74_0 (this is a generated file);
print('===== testing package: thinc-8.0.1-py37h1341a74_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import os
import sys
import pytest
import thinc

PACKAGE_DIR = os.path.abspath(os.path.dirname(thinc.__file__))
sys.exit(pytest.main([PACKAGE_DIR]))
#  --- run_test.py (end) ---

print('===== thinc-8.0.1-py37h1341a74_0 OK =====');
print("import: 'thinc'")
import thinc

print("import: 'thinc.api'")
import thinc.api

print("import: 'thinc.linear'")
import thinc.linear

print("import: 'thinc.extra'")
import thinc.extra

print("import: 'thinc.neural'")
import thinc.neural

